@@ -391,33 +391,14 @@ extern const enum reg_class picochip_regno_reg_class[FIRST_PSEUDO_REGISTER];
 #define FRAME_POINTER_REGNUM 18
 #define ARG_POINTER_REGNUM   19
 
-/* Static chain is used to pass the local variables of the enclosing function.
-   The static chain is passed in memory. The first long-word location
-   beneath the stack pointer is used. In the presence of pretend
-   arguments, which are written into that location, this mechanism
-   complicates matters. */
-
-/* Location seen by the caller. */
-#define STATIC_CHAIN							\
-  gen_rtx_MEM (Pmode, plus_constant (stack_pointer_rtx, -2 * UNITS_PER_WORD))
-
-/* Location seen by the callee. */
-#define STATIC_CHAIN_INCOMING						\
-  gen_rtx_MEM (Pmode, plus_constant (arg_pointer_rtx, 0))
-
 /* Eliminating Frame Pointer and Arg Pointer.  The frame and argument
    pointers are eliminated wherever possible, by replacing them with
    offsets from the stack pointer. */
 
-/* We want to get rid of the frame pointer.  */
-#define FRAME_POINTER_REQUIRED 0
-
 #define ELIMINABLE_REGS 						\
   {{ARG_POINTER_REGNUM, STACK_POINTER_REGNUM},				\
    {FRAME_POINTER_REGNUM, STACK_POINTER_REGNUM}}
 
-#define CAN_ELIMINATE(FROM, TO) 1
-
 #define INITIAL_ELIMINATION_OFFSET(FROM,TO,OFFSET) \
   OFFSET = initial_elimination_offset(FROM, TO);
 
@@ -487,30 +468,11 @@ extern const enum reg_class picochip_regno_reg_class[FIRST_PSEUDO_REGISTER];
 
 /* No trampolines.  */
 #define TRAMPOLINE_SIZE 0
-#define INITIALIZE_TRAMPOLINE(ADDR,FNADDR,CHAIN)
 
 /* Addressing Modes  */
 
-#define CONSTANT_ADDRESS_P(X) CONSTANT_P(X)
-
 #define MAX_REGS_PER_ADDRESS 1
 
-#ifdef REG_OK_STRICT
-
-#define GO_IF_LEGITIMATE_ADDRESS(MODE, X, LABEL) 			\
- if (picochip_legitimate_address_p (MODE, X, 1)) goto LABEL;
-
-#else /* REG_OK_STRICT */
-
-#define GO_IF_LEGITIMATE_ADDRESS(MODE, X, LABEL) 			\
-  if (picochip_legitimate_address_p (MODE, X, 0)) goto LABEL;
-
-#endif /* !REG_OK_STRICT */
-
-/* extern struct rtx_def *picochip_legitimize_address */
-/* 	PARAMS ((struct rtx_def *, struct rtx_def *, int)); */
-#define LEGITIMIZE_ADDRESS(X,OLDX,MODE,WIN);
-
 /* Legitimize reload address tries machine dependent means of
    reloading addresses.  There seems to be a strange error in gcc,
    which necessitates this macro.  Consider:
@@ -535,9 +497,6 @@ extern const enum reg_class picochip_regno_reg_class[FIRST_PSEUDO_REGISTER];
 #define LEGITIMIZE_RELOAD_ADDRESS(X,MODE,OPNUM,TYPE,IND_LEVELS,WIN)	     \
 if (picochip_symbol_offset(X)) { X = gen_rtx_CONST(MODE, X); }
 
-/* There are no mode dependent addresses.  */
-#define GO_IF_MODE_DEPENDENT_ADDRESS(ADDR,LABEL) do {} while (0)
-
 /* Nonzero if the constant rtx X is a legitimate general operand.  X
    satisfies CONSTANT_P.  */
 
@@ -739,6 +698,7 @@ enum picochip_builtins
   PICOCHIP_BUILTIN_HALT
 };
 
+#define NO_DOLLAR_IN_LABEL 1
 #define NO_DOT_IN_LABEL 1
 
 /* The assembler does support LEB128, despite the auto-configure test
